
(load "elems.scm")	; vyhodnot obsah souboru elems.scm

;;;;
;;;;  READER (nacteni vyrazu a prevedeni do interni representace)
;;;;

;; prevedeni vyrazu do interni formy
(define expr->intern
  (lambda (expr)
    (cond ((symbol? expr) (make-symbol expr))
	  ((number? expr) (make-number expr))
	  ((and (boolean? expr) expr) scm-true)
	  ((boolean? expr) scm-false)
	  ((null? expr) the-empty-list)
	  ((pair? expr) (make-pair (expr->intern (car expr))
				   (expr->intern (cdr expr))))
	  ((eof-object? expr) #f)
	  (else (error "; Syntactic error.")))))

;; nacti vstupni vyraz do interni formy
(define scm-read
  (lambda ()
    (expr->intern (read))))

;;;;
;;;;  PRINTER (vytisteni vyrazu v jeho externi reprez. -- pokud je to mozne)
;;;;

;; pouze pouzije display -- vypise syrovou reprezentaci
(define scm-print
  (lambda (elem)
    (display elem)))

;; prevedeni elementu do externi formy
(define intern->expr
  (lambda (expr)
    (cond ((scm-pair? expr) (cons (intern->expr (pair-car expr))
				  (intern->expr (pair-cdr expr))))
	  ((scm-env? expr) "#<environment>")
	  ((scm-primitive? expr) "#<primitive-procedure>")
	  ((scm-user-procedure? expr) "#<user-defined-procedure>")
	  ((scm-macro? expr) "#<macro>")
	  ((scm-generated-symbol? expr) "#<generated-symbol>")
	  ((scm-specform? expr) "#<special-form>")
	  ((scm-undefined? expr) "#<undefined>")
	  ((scm-void? expr) "#<void>")
	  (else (get-data expr)))))

;; o neco ,,chytrejsi printer''
(define scm-print
  (lambda (elem)
    (display (intern->expr elem))))

;;;;
;;;;  EVALUATOR
;;;;

;; map pres elementy tvorici scm-seznam, vysledkem je klasicky seznam
(define map-scm-list->list
  (lambda (f scm-l)
    (convert-list scm-null? pair-car pair-cdr cons '() f scm-l)))

;; preved scm-seznam na klasicky seznam
(define scm-list->list
  (lambda (scm-l)
    (convert-list scm-null? pair-car pair-cdr cons '() (lambda (x) x) scm-l)))

;; preved klasicky seznam na scm-seznam
(define list->scm-list
  (lambda (l)
    (convert-list null? car cdr make-pair the-empty-list (lambda (x) x) l)))

;; vyhodnot vyraz v danem prostredi
(define scm-eval
  (lambda (elem env)

    ;; vyhodnocovani elementu podle jejich typu
    (cond

     ;; symboly se vyhodnocuji na svou aktualni vazbu
     ((scm-symbol? elem)
      (let* ((binding (lookup-env env elem #t #f)))
	(if binding
	    (pair-cdr binding)
	    (begin
	      (display elem)
	    (error "; EVAL: Symbol not bound")))))

     ;; vyhodnoceni seznamu
     ((scm-pair? elem)

      ;; nejprve vyhodnotime prvni prvek seznamu
      (let* ((first (pair-car elem))
	     (args (pair-cdr elem))
	     (f (scm-eval first env)))

	;; podle prvniho prvku rozhodni o co se jedna
	(cond

	 ;; pokud se jedna o proceduru, vyhodnot argumenty a aplikuj
	 ((scm-procedure? f)
	  (scm-apply f (map-scm-list->list
			(lambda (elem)
			  (scm-eval elem env))
			args)))

	 ;; pokud se jedna o formu, aplikuj s nevyhodnocenymi argumenty
	 ((scm-form? f)
	  (scm-form-apply env f (scm-list->list args)))
	 
	 ;; na prvnim miste stoji nepripustny prvek
	 (error "; EVAL: First element did not eval. to procedure"))))

     ;; vse ostatni (cisla, boolean, ... se vyhodnocuje na sebe sama)
     (else elem))))

;; vytvor tabulku vazeb: formalni argument -- argument
(define make-bindings
  (lambda (formal-args args)
    (cond ((scm-null? formal-args) the-empty-list)
	  ((scm-symbol? formal-args)
	   (make-pair (make-pair formal-args (list->scm-list args))
		      the-empty-list))
	  (else (make-pair
		 (make-pair (pair-car formal-args) (car args))
		 (make-bindings (pair-cdr formal-args) (cdr args)))))))

;; aplikuj proceduru, predka nastav na ``env''
(define scm-env-apply
  (lambda (proc env args)
    (cond ((scm-primitive? proc) (apply (get-data proc) args))
	  ((scm-user-procedure? proc)
	   (scm-eval (procedure-body proc)
		     (make-env env
			       (make-bindings (procedure-arguments proc)
					      args))))
	  (else (error "; APPLY: Expected procedure")))))

;; aplikuj proceduru s lexikalnim predkem
(define scm-apply
  (lambda (proc args)
    (cond ((scm-primitive? proc) (scm-env-apply proc #f args))
	  ((scm-user-procedure? proc)
	   (scm-env-apply proc (procedure-environment proc) args))
	  (else (error "; APPLY: Expected procedure")))))

;; aplikuj spec. formu
(define scm-form-apply
  (lambda (env form args)
    (cond ((scm-specform? form) (apply (get-data form) env args))
	  ((scm-macro? form) (scm-eval (scm-apply (get-data form) args) env))
	  (else (error "; APPLY: Expected special form")))))

;;;;
;;;;  NOVY REPL
;;;;

(load "globenv.scm")	; vyhodnot obsah souboru globenv.scm

;; procedura load: vyhodnoceni vyrazu v souboru (slouzi k natazeni knihovny)
(define scm-load
  (lambda (file-name env)
    (let ((port (open-input-file file-name)))
      (let next ()
	(let ((expr (read port)))
	  (if (not (eof-object? expr))
	      (begin
		(scm-eval (expr->intern expr) env)
		(next)))))
      (close-input-port port))))

;; vrat nazev typu jako retezec s prvnim pismenem velkym
(define type-name
  (lambda (elem)
    (let ((strl (string->list (symbol->string (get-type-tag elem)))))
      (list->string (cons (char-upcase (car strl)) (cdr strl))))))

;; cyklus REPL
(define scm-repl
  (lambda ()
    (let ((glob-env (make-env scheme-toplevel-env the-empty-list)))
      (scm-load "includes.scm" glob-env)
      (let loop ()
	(display "]=> ")
	(let ((elem (scm-read)))
	  (if (not elem)
	      'bye-bye
	      (let ((result (scm-eval elem glob-env)))
		(newline)
		(display (string-append "; " (type-name result)	": "))
		(scm-print result)
		(newline)
		(newline)
		(loop))))))))

;; spust REPL
(scm-repl)
